"""Tests for lakshmi.utils module."""
import unittest

import yaml

from lakshmi import utils


class UtilsTest(unittest.TestCase):
    def test_format_money(self):
        self.assertEqual('$42.42', utils.format_money(42.421))
        self.assertEqual('$100.00', utils.format_money(100.00))

    def test_format_money_delta(self):
        self.assertEqual('+$10.00', utils.format_money_delta(10))
        self.assertEqual('-$20.05', utils.format_money_delta(-20.049))

    def test_validate_date_no_error(self):
        self.assertEqual('2021/11/21', utils.validate_date('2021/11/21'))

    def test_validate_date_corrected(self):
        self.assertEqual('2020/01/01', utils.validate_date('2020/1/1'))

    def test_validate_date_errors(self):
        with self.assertRaises(ValueError):
            utils.validate_date('01/23/2021')

        with self.assertRaises(ValueError):
            utils.validate_date('2021/02/29')  # 2021 is not leap year.

    def test_resolver(self):
        data = 'a: 100.22\nb: 122,121,000.22\nc: -121,122.12\nd: 1,000'
        loaded = yaml.load(data, Loader=utils.get_loader())
        self.assertEqual(100.22, loaded['a'])
        self.assertEqual(122121000.22, loaded['b'])
        self.assertEqual(-121122.12, loaded['c'])
        self.assertEqual(1000, loaded['d'])



"""Tests for lakshmi.cache module."""
import pickle
import unittest
from pathlib import Path
from unittest.mock import Mock, patch

import lakshmi.cache as cache


class Cached(cache.Cacheable):
    def __init__(self, key, value):
        self.key = key
        self.value = value

    def cache_key(self):
        return self.key

    @cache.cache(2)
    def get_value(self):
        return self.value

    @cache.cache(365)
    def get_long_cached_value(self):
        return self.value


class CacheTest(unittest.TestCase):
    def setUp(self):
        # Reset cache dir setting.
        cache._prefetch_obj = None
        cache._ctx.pop(cache._CACHE_STR, None)
        cache.set_force_refresh(False)
        cache.set_cache_miss_func(None)

    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    def test_disabled_cache(self, get_file_age, exists):
        cache.set_cache_dir(None)  # Disble caching.
        c = Cached('key1', 1)
        self.assertEqual(1, c.get_value())
        c.value = 2
        self.assertEqual(2, c.get_value())
        get_file_age.assert_not_called()
        exists.assert_not_called()

    def test_disabled_cache_with_func(self):
        cache.set_cache_dir(None)  # Disble caching.
        mocked_obj = Mock()
        cache.set_cache_miss_func(mocked_obj.func)

        c = Cached('key1', 1)
        self.assertEqual(1, c.get_value())
        mocked_obj.func.assert_called_once()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_default_cache_miss(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        def side_effect(x):
            cache._ctx[cache._CACHE_STR] = x
        set_cache_dir.side_effect = side_effect
        exists.return_value = False

        c = Cached('key1', 1)
        self.assertEqual(1, c.get_value())

        set_cache_dir.assert_called_once()
        get_file_age.assert_not_called()
        exists.assert_called_once()
        write_bytes.assert_called_once_with(pickle.dumps(1))
        read_bytes.assert_not_called()

    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.set_cache_dir')
    def test_default_cache_miss_with_func(
            self, set_cache_dir, exists, write_bytes):
        def side_effect(x):
            cache._ctx[cache._CACHE_STR] = x
        set_cache_dir.side_effect = side_effect
        exists.return_value = False
        mocked_obj = Mock()
        cache.set_cache_miss_func(mocked_obj.func)

        c = Cached('key1', 1)
        self.assertEqual(1, c.get_value())

        set_cache_dir.assert_called_once()
        exists.assert_called_once()
        write_bytes.assert_called_once_with(pickle.dumps(1))
        mocked_obj.func.assert_called_once()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_default_cache_hit(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        def side_effect(x):
            cache._ctx[cache._CACHE_STR] = x
        set_cache_dir.side_effect = side_effect
        exists.return_value = True
        get_file_age.return_value = 1
        read_bytes.return_value = pickle.dumps(1)  # Cache 1.

        c = Cached('key2', 2)
        self.assertEqual(1, c.get_value())  # Cached value.

        set_cache_dir.assert_called_once()
        get_file_age.assert_called_once()
        exists.assert_called_once()
        write_bytes.assert_not_called()
        read_bytes.assert_called_once()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_default_cache_hit_with_func(
            self, set_cache_dir, get_file_age, exists, read_bytes):
        def side_effect(x):
            cache._ctx[cache._CACHE_STR] = x
        set_cache_dir.side_effect = side_effect
        exists.return_value = True
        get_file_age.return_value = 1
        read_bytes.return_value = pickle.dumps(1)  # Cache 1.
        mocked_obj = Mock()
        cache.set_cache_miss_func(mocked_obj.func)

        c = Cached('key2', 2)
        self.assertEqual(1, c.get_value())  # Cached value.

        set_cache_dir.assert_called_once()
        get_file_age.assert_called_once()
        exists.assert_called_once()
        read_bytes.assert_called_once()
        mocked_obj.func.assert_not_called()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_set_cache(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        cache._ctx[cache._CACHE_STR] = Path('/fake/dir')
        exists.return_value = False

        c = Cached('key1', 1)
        self.assertEqual(1, c.get_value())

        set_cache_dir.assert_not_called()
        get_file_age.assert_not_called()
        exists.assert_called_once()
        write_bytes.assert_called_once_with(pickle.dumps(1))
        read_bytes.assert_not_called()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_force_refresh(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        cache._ctx[cache._CACHE_STR] = Path('/fake/dir')
        cache.set_force_refresh(True)

        c = Cached('key2', 2)
        self.assertEqual(2, c.get_value())  # Cached value not used.

        set_cache_dir.assert_not_called()
        get_file_age.assert_not_called()
        exists.assert_not_called()
        write_bytes.assert_called_once_with(pickle.dumps(2))
        read_bytes.assert_not_called()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_force_refresh_with_long_cached_value(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        cache._ctx[cache._CACHE_STR] = Path('/fake/dir')
        cache.set_force_refresh(True)
        exists.return_value = True
        get_file_age.return_value = 5
        read_bytes.return_value = pickle.dumps(1)  # Cache 1.

        c = Cached('key2', 2)
        self.assertEqual(1, c.get_long_cached_value())  # Cached value.

        set_cache_dir.assert_not_called()
        get_file_age.assert_called_once()
        exists.assert_called_once()
        write_bytes.assert_not_called()
        read_bytes.assert_called_once()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_force_refresh_called_twice(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        # While this test is not logically "correct", it makes the
        # testing easier and more robust. In reality the cached
        # value would be 2 and not 3.
        cache._ctx[cache._CACHE_STR] = Path('/fake/dir')
        cache.set_force_refresh(True)
        get_file_age.return_value = 1
        read_bytes.return_value = pickle.dumps(3)  # Cache 3.

        c = Cached('key2', 2)
        self.assertEqual(2, c.get_value())  # Cached value not used.
        self.assertEqual(3, c.get_value())  # Cached value used.

        set_cache_dir.assert_not_called()
        get_file_age.assert_called_once()
        exists.assert_called_once()
        write_bytes.assert_called_once_with(pickle.dumps(2))
        read_bytes.assert_called_once()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_old_cache(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        cache._ctx[cache._CACHE_STR] = Path('/fake/dir')
        exists.return_value = True
        get_file_age.return_value = 2

        c = Cached('key1', 1)
        self.assertEqual(1, c.get_value())

        set_cache_dir.assert_not_called()
        get_file_age.assert_called_once()
        exists.assert_called_once()
        write_bytes.assert_called_once_with(pickle.dumps(1))
        read_bytes.assert_not_called()

    @patch('lakshmi.cache.set_cache_dir')
    def test_return_cached_funcs(self, set_cache_dir):
        def side_effect(x):
            cache._ctx[cache._CACHE_STR] = x
        set_cache_dir.side_effect = side_effect
        prefetch = cache._Prefetch()
        c = Cached('key1', 1)
        self.assertEqual(2, len(prefetch._return_cached_funcs(c)))

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_cache_miss_prefetch(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        def side_effect(x):
            cache._ctx[cache._CACHE_STR] = x
        set_cache_dir.side_effect = side_effect
        exists.return_value = False

        c = Cached('key1', 1)
        cache.prefetch_add(c)
        cache.prefetch()

        set_cache_dir.assert_called_once()
        get_file_age.assert_not_called()
        exists.assert_called()
        self.assertEqual(2, write_bytes.call_count)
        read_bytes.assert_not_called()

    @patch('pathlib.Path.read_bytes')
    @patch('pathlib.Path.write_bytes')
    @patch('pathlib.Path.exists')
    @patch('lakshmi.cache.get_file_age')
    @patch('lakshmi.cache.set_cache_dir')
    def test_force_refresh_with_prefetch(
            self, set_cache_dir, get_file_age, exists, write_bytes,
            read_bytes):
        cache._ctx[cache._CACHE_STR] = Path('/fake/dir')
        cache.set_force_refresh(True)
        get_file_age.return_value = 2

        c = Cached('key2', 9)
        cache.prefetch_add(c)
        cache.prefetch()
        self.assertEqual(9, c.get_value())

        set_cache_dir.assert_not_called()
        self.assertEqual(2, get_file_age.call_count)
        self.assertEqual(2, exists.call_count)
        self.assertEqual(2, write_bytes.call_count)
        read_bytes.assert_not_called()

    def test_prefetch_add_not_called(self):
        # Do not fail if there are no functions to be prefetched.
        cache.prefetch()


if __name__ == '__main__':
    unittest.main()



"""Tests for lakshmi/data directory. Simply checks if the files parses."""
import unittest
from pathlib import Path
from unittest.mock import patch

import ibonds
import yaml

import lakshmi
from lakshmi import utils


class DataTest(unittest.TestCase):
    def parse_dict(self, filename, function):
        file_path = (Path(__file__).parents[1].absolute() / filename)
        d = yaml.load(file_path.read_text(), Loader=utils.get_loader())
        return function(d)

    def test_account(self):
        self.assertIsNotNone(self.parse_dict('lakshmi/data/Account.yaml',
                                             lakshmi.Account.from_dict))

    def test_asset_class(self):
        self.assertIsNotNone(self.parse_dict('lakshmi/data/AssetClass.yaml',
                                             lakshmi.AssetClass.from_dict))

    def test_ee_bonds(self):
        self.assertIsNotNone(self.parse_dict('lakshmi/data/EEBonds.yaml',
                                             lakshmi.assets.EEBonds.from_dict))

    @patch('lakshmi.assets.IBonds._InterestRates.get')
    def test_i_bonds(self, mock_get):
        INTEREST_RATE_DATA = """
        2020-11-01:
        - 0.00
        - 0.84
        2021-05-01:
        - 0.00
        - 1.77
        """
        mock_get.return_value = ibonds.InterestRates(INTEREST_RATE_DATA)
        self.assertIsNotNone(self.parse_dict('lakshmi/data/IBonds.yaml',
                                             lakshmi.assets.IBonds.from_dict))

    def test_manual_asset(self):
        self.assertIsNotNone(
            self.parse_dict(
                'lakshmi/data/ManualAsset.yaml',
                lakshmi.assets.ManualAsset.from_dict))

    def test_ticker_asset(self):
        self.assertIsNotNone(
            self.parse_dict(
                'lakshmi/data/TickerAsset.yaml',
                lakshmi.assets.TickerAsset.from_dict))

    def test_vanguard_fund(self):
        self.assertIsNotNone(
            self.parse_dict(
                'lakshmi/data/VanguardFund.yaml',
                lakshmi.assets.VanguardFund.from_dict))

    def test_checkpoint(self):
        self.assertIsNotNone(
            self.parse_dict(
                'lakshmi/data/Checkpoint.yaml',
                lambda x: lakshmi.performance.Checkpoint.from_dict(
                    x, date='2021/01/01')))

    def test_portfolio(self):
        self.assertIsNotNone(self.parse_dict('docs/portfolio.yaml',
                                             lakshmi.Portfolio.from_dict))


if __name__ == '__main__':
    unittest.main()



"""Tests for lakshmi.assets module."""
import datetime
import json
import pathlib
import unittest
from unittest.mock import MagicMock, patch

import ibonds

import lakshmi.assets as assets
import lakshmi.cache


class AssetsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        lakshmi.cache.set_cache_dir(None)  # Disable caching.
        cls.data_dir = (pathlib.Path(__file__).parent / 'data')

    def test_dict_manual_asset_with_what_if(self):
        manual_asset = assets.ManualAsset('Cash', 100.5, {'Fixed Income': 1.0})
        manual_asset.what_if(100)
        manual_asset = assets.from_dict(assets.to_dict(manual_asset))
        self.assertEqual('Cash', manual_asset.name())
        self.assertAlmostEqual(100.5, manual_asset.value())
        self.assertAlmostEqual(200.5, manual_asset.adjusted_value())
        self.assertEqual({'Fixed Income': 1.0}, manual_asset.class2ratio)

    def test_asset_bad_what_if(self):
        a = assets.ManualAsset('Cash', 100, {'All': 1.0})
        a.what_if(-101)
        self.assertAlmostEqual(0, a.adjusted_value())

    def test_dict_manual_asset(self):
        manual_asset = assets.ManualAsset('Cash', 100.5, {'Fixed Income': 1.0})
        manual_asset = assets.from_dict(assets.to_dict(manual_asset))
        self.assertEqual('Cash', manual_asset.name())
        self.assertAlmostEqual(100.5, manual_asset.adjusted_value())
        self.assertEqual({'Fixed Income': 1.0}, manual_asset.class2ratio)

    def test_manual_asset_to_table(self):
        manual_asset = assets.ManualAsset('Cash', 100.5, {'Fixed Income': 1.0})
        expected = [['Name:', 'Cash'],
                    ['Asset Class Mapping:', 'Fixed Income  100%'],
                    ['Value:', '$100.50']]
        self.assertListEqual(expected, manual_asset.to_table().str_list())

        manual_asset.what_if(-100)
        expected = [['Name:', 'Cash'],
                    ['Asset Class Mapping:', 'Fixed Income  100%'],
                    ['Adjusted Value:', '$0.50'],
                    ['What if:', '-$100.00']]
        self.assertListEqual(expected, manual_asset.to_table().str_list())

    @patch('yfinance.Ticker')
    def test_bad_ticker(self, MockTicker):
        bad_ticker = MagicMock()
        bad_ticker.info = {}
        bad_ticker.fast_info = {}
        MockTicker.return_value = bad_ticker

        ticker_asset = assets.TickerAsset('bad', 10, {'All': 1.0})
        with self.assertRaisesRegex(assets.NotFoundError,
                                    'Cannot retrieve ticker'):
            ticker_asset.name()
        with self.assertRaisesRegex(assets.NotFoundError,
                                    'Cannot retrieve ticker'):
            ticker_asset.value()

        MockTicker.assert_called_once()

    @patch('yfinance.Ticker')
    def test_good_ticker(self, MockTicker):
        ticker = MagicMock()
        ticker.info = {'longName': 'Vanguard Cash Reserves Federal'}
        ticker.fast_info = {'lastPrice': 1.0}
        MockTicker.return_value = ticker

        vmmxx = assets.TickerAsset('VMMXX', 100.0, {'All': 1.0})
        self.assertAlmostEqual(100.0, vmmxx.value())
        self.assertEqual('Vanguard Cash Reserves Federal', vmmxx.name())
        self.assertEqual('VMMXX', vmmxx.short_name())

        MockTicker.assert_called_once()

    @patch('yfinance.Ticker')
    def test_missing_longname(self, MockTicker):
        ticker = MagicMock()
        ticker.info = {'shortName': 'Bitcoin USD',
                       'name': 'Bitcoin'}
        MockTicker.return_value = ticker

        btc = assets.TickerAsset('BTC-USD', 1.0, {'All': 1.0})
        self.assertEqual('Bitcoin USD', btc.name())

        MockTicker.assert_called_once()

    @patch('yfinance.Ticker')
    def test_missing_longname_shortname(self, MockTicker):
        ticker = MagicMock()
        ticker.info = {'name': 'Bitcoin'}
        ticker.fast_info = {'regularMarketPrice': 1.0}
        MockTicker.return_value = ticker

        btc = assets.TickerAsset('BTC-USD', 1.0, {'All': 1.0})
        self.assertEqual('Bitcoin', btc.name())

        MockTicker.assert_called_once()

    def test_tax_lots_ticker(self):
        vmmxx = assets.TickerAsset('VMMXX', 100.0, {'All': 1.0})
        lots = [assets.TaxLot('2012/12/12', 50, 1.0),
                assets.TaxLot('2013/12/12', 30, 0.9)]
        with self.assertRaisesRegex(AssertionError,
                                    'Lots provided should sum up to 100.0'):
            vmmxx.set_lots(lots)

        lots.append(assets.TaxLot('2014/12/31', 20, 0.9))
        vmmxx.set_lots(lots)
        self.assertListEqual(lots, vmmxx.get_lots())

    @patch('lakshmi.assets.TickerAsset.price')
    def test_list_lots(self, mock_price):
        mock_price.return_value = 15.0

        vti = assets.TickerAsset('VTI', 100.0, {'All': 1.0})
        lots = [assets.TaxLot('2011/01/01', 50, 10.0),
                assets.TaxLot('2012/01/01', 50, 20.0)]
        vti.set_lots(lots)

        self.assertListEqual(
            [['2011/01/01', '50.0', '$500.00', '+$250.00', '50.0%'],
             ['2012/01/01', '50.0', '$1,000.00', '-$250.00', '-25.0%']],
            vti.list_lots().str_list())

    @patch('lakshmi.assets._today')
    @patch('lakshmi.assets.TickerAsset.price')
    def test_list_lots_with_term(self, mock_price, mock_today):
        mock_price.return_value = 15.0
        mock_today.return_value = datetime.datetime.strptime(
            '2012/12/01', '%Y/%m/%d')

        vti = assets.TickerAsset('VTI', 150.0, {'All': 1.0})
        lots = [assets.TaxLot('2011/01/01', 50, 10.0),
                assets.TaxLot('2012/01/01', 50, 20.0),
                assets.TaxLot('2012/11/01', 50, 20.0)]
        vti.set_lots(lots)

        self.assertListEqual(
            [['2011/01/01', '50.0', '$500.00', '+$250.00', '50.0%', 'LT'],
             ['2012/01/01', '50.0', '$1,000.00', '-$250.00', '-25.0%', 'ST'],
             ['2012/11/01', '50.0', '$1,000.00', '-$250.00', '-25.0%', '30']],
            vti.list_lots(include_term=True).str_list())

    @patch('lakshmi.assets.TickerAsset.name')
    @patch('lakshmi.assets.TickerAsset.price')
    def test_ticker_asset_to_table(self, mock_price, mock_name):
        mock_price.return_value = 10.0
        mock_name.return_value = 'Google Inc'

        goog = assets.TickerAsset('GOOG', 100.0, {'All': 1.0})
        expected = [['Ticker:', 'GOOG'],
                    ['Name:', 'Google Inc'],
                    ['Asset Class Mapping:', 'All  100%'],
                    ['Value:', '$1,000.00'],
                    ['Price:', '$10.00']]
        self.assertListEqual(expected, goog.to_table().str_list())

    @patch('yfinance.Ticker')
    def test_dict_ticker(self, MockTicker):
        ticker = MagicMock()
        ticker.info = {'longName': 'Vanguard Cash Reserves Federal'}
        ticker.fast_info = {'lastPrice': 1.0}
        MockTicker.return_value = ticker

        vmmxx = assets.TickerAsset('VMMXX', 100.0, {'All': 1.0})
        lots = [assets.TaxLot('2012/12/12', 50, 1.0),
                assets.TaxLot('2013/12/12', 50, 0.9)]
        vmmxx.set_lots(lots)
        vmmxx.what_if(-10)
        vmmxx = assets.from_dict(assets.to_dict(vmmxx))
        self.assertEqual('VMMXX', vmmxx.short_name())
        self.assertEqual(100.0, vmmxx.shares())
        self.assertEqual({'All': 1.0}, vmmxx.class2ratio)
        self.assertAlmostEqual(90.0, vmmxx.adjusted_value())
        self.assertEqual(2, len(vmmxx.get_lots()))

    @patch('requests.Session.get')
    def test_vanguard_funds_name(self, mock_get):
        mock_res = MagicMock()

        with open(self.data_dir / 'profile.json') as data_file:
            mock_res.json.return_value = json.load(data_file)

        mock_get.return_value = mock_res

        fund = assets.VanguardFund(7555, 10, {'All': 1.0})
        self.assertEqual(
            'Vanguard Institutional Total Bond Market Index Trust',
            fund.name())
        self.assertEqual('7555', fund.short_name())
        mock_get.assert_called_once_with(
            'https://api.vanguard.com/rs/ire/01/pe/fund/7555/profile.json',
            headers={'Referer': 'https://vanguard.com/'}, verify=False)

    @patch('requests.Session.get')
    def test_vanguard_funds_value(self, mock_get):
        mock_res = MagicMock()

        with open(self.data_dir / 'price.json') as data_file:
            mock_res.json.return_value = json.load(data_file)
        mock_get.return_value = mock_res

        fund = assets.VanguardFund(7555, 10, {'All': 1.0})
        self.assertEqual(1166.6, fund.value())
        mock_get.assert_called_once_with(
            'https://api.vanguard.com/rs/ire/01/pe/fund/7555/price.json',
            headers={'Referer': 'https://vanguard.com/'}, verify=False)
        fund.set_lots([assets.TaxLot('2012/12/30', 10, 1.0)])

    @patch('lakshmi.assets.VanguardFund.value')
    def test_dict_vanguard_fund(self, mock_value):
        mock_value.return_value = 100.0
        fund = assets.VanguardFund(1234, 20, {'Bonds': 1.0})
        fund.set_lots([assets.TaxLot('2021/05/15', 20, 5.0)])
        fund.what_if(100)
        fund = assets.from_dict(assets.to_dict(fund))
        self.assertEqual('1234', fund.short_name())
        self.assertEqual(20, fund.shares())
        self.assertEqual({'Bonds': 1.0}, fund.class2ratio)
        self.assertEqual(1, len(fund.get_lots()))
        self.assertEqual(100, fund._delta)

    @patch('lakshmi.assets.VanguardFund.name')
    @patch('lakshmi.assets.VanguardFund.price')
    def test_vangurd_fund_asset_to_table(self, mock_price, mock_name):
        mock_price.return_value = 10.0
        mock_name.return_value = 'Vanguardy Fund'

        fund = assets.VanguardFund(123, 100.0, {'All': 1.0})
        expected = [['Fund id:', '123'],
                    ['Name:', 'Vanguardy Fund'],
                    ['Asset Class Mapping:', 'All  100%'],
                    ['Value:', '$1,000.00'],
                    ['Price:', '$10.00']]
        self.assertListEqual(expected, fund.to_table().str_list())

    @patch('lakshmi.assets._today_date')
    @patch('lakshmi.assets.IBonds._InterestRates.get')
    def test_i_bonds(self, mock_get, mock_today):
        INTEREST_RATE_DATA = """
        2020-11-01:
        - 0.00
        - 0.84
        2021-05-01:
        - 0.00
        - 1.77
        """
        mock_get.return_value = ibonds.InterestRates(INTEREST_RATE_DATA)
        ibond_asset = assets.IBonds({'All': 1.0})
        ibond_asset.add_bond('11/2020', 10000)

        self.assertEqual('I Bonds', ibond_asset.name())
        self.assertEqual('I Bonds', ibond_asset.short_name())

        mock_today.return_value = datetime.date(2020, 11, 2)
        self.assertListEqual(
            [['11/2020', '$10,000.00', '0.00%', '1.68%', '$10,000.00']],
            ibond_asset.list_bonds().str_list())

        # Test the case where the interest rate data is not up to date.
        mock_today.return_value = datetime.date(2021, 11, 1)
        self.assertListEqual(
            [['11/2020', '$10,000.00', '0.00%', '', '$10,172.00']],
            ibond_asset.list_bonds().str_list())

    @patch('lakshmi.assets.IBonds.value')
    def test_dict_i_bonds(self, mock_value):
        mock_value.return_value = 11000
        ibonds = assets.IBonds({'B': 1.0})
        ibonds.add_bond('02/2020', 10000)
        ibonds.what_if(-100.0)

        ibonds = assets.from_dict(assets.to_dict(ibonds))
        self.assertEqual('I Bonds', ibonds.name())
        self.assertEqual({'B': 1.0}, ibonds.class2ratio)
        self.assertAlmostEqual(-100.0, ibonds._delta)
        self.assertEqual(1, len(ibonds.bonds()))

    @patch('datetime.datetime')
    @patch('requests.post')
    def test_ee_bonds(self, mock_post, mock_date):
        mock_res = MagicMock()
        with open(self.data_dir / 'SBCPrice-EE.html') as html_file:
            mock_res.text = html_file.read()
        mock_post.return_value = mock_res
        mock_date.now.strftime.return_value = '04/2021'
        # Bypass issue date validation.
        mock_strptime = MagicMock()
        mock_strptime.strftime.return_value = '03/2020'
        mock_date.strptime.return_value = mock_strptime

        eebonds = assets.EEBonds({'All': 1.0})
        eebonds.add_bond('3/2020', 10000)

        mock_post.asset_called_once_with(
            'http://www.treasurydirect.gov/BC/SBCPrice',
            data={
                'RedemptionDate': '04/2021',
                'Series': 'EE',
                'Denomination': '500',
                'IssueDate': '03/2020',
                'btnAdd.x': 'CALCULATE'})

        self.assertEqual('EE Bonds', eebonds.name())
        self.assertEqual('EE Bonds', eebonds.short_name())
        self.assertAlmostEqual(10008.0, eebonds.value())
        self.assertListEqual(
            [['03/2020', '$10,000.00', '0.10%', '$10,008.00']],
            eebonds.list_bonds().str_list())

    @patch('lakshmi.assets.EEBonds.value')
    def test_dict_ee_bonds(self, mock_value):
        mock_value.return_value = 10010
        eebonds = assets.EEBonds({'B': 1.0})
        eebonds.add_bond('02/2020', 10000)
        eebonds.what_if(-100.0)
        eebonds = assets.from_dict(assets.to_dict(eebonds))
        self.assertEqual('EE Bonds', eebonds.name())
        self.assertEqual({'B': 1.0}, eebonds.class2ratio)
        self.assertAlmostEqual(-100.0, eebonds._delta)
        self.assertEqual(1, len(eebonds.bonds()))


if __name__ == '__main__':
    unittest.main()



"""Tests for lakshmi.lak application."""
import unittest
from pathlib import Path
from unittest.mock import patch

import click
from click.testing import CliRunner

from lakshmi import Account, AssetClass, Portfolio, lak
from lakshmi.assets import ManualAsset
from lakshmi.performance import Checkpoint, Performance, Timeline


class TestLakContext(lak.LakContext):
    """A testing version of LakContext that doesn't load or save
    portfolio."""

    def __init__(self):
        self.portfolio_filename = 'test_portfolio.yaml'
        self.performance_filename = 'test_performance.yaml'
        self.continued = False
        self.warned = False
        self.whatifs = None
        self.tablefmt = None
        self.saved_portfolio = False
        self.saved_performance = False

        self.portfolio = Portfolio(
            AssetClass('All')
            .add_subclass(0.5, AssetClass('Stocks'))
            .add_subclass(0.5, AssetClass('Bonds'))).add_account(
            Account('Schwab', 'Taxable').add_asset(
                ManualAsset('Test Asset', 100.0, {'Stocks': 1.0})))

        self.performance = Performance(Timeline([
            Checkpoint('2021/1/1', 100),
            Checkpoint('2021/1/2', 105.01, inflow=10, outflow=5)]))

    def get_portfolio(self):
        return self.portfolio

    def save_portfolio(self):
        self.saved_portfolio = True

    def get_performance(self):
        if self.performance:
            return self.performance
        raise click.ClickException('Performance file not found.')

    def save_performance(self):
        self.saved_performance = True

    def reset(self):
        """Reset the state for a new command (except the portfolio)."""
        self.continued = False
        self.warned = False
        self.whatifs = None
        self.tablefmt = None
        self.saved_portfolio = False


def run_lak(args):
    return CliRunner().invoke(lak.lak, args.split(' '))


class LakTest(unittest.TestCase):
    def setUp(self):
        lak.lakctx = TestLakContext()

    @patch('lakshmi.lak.LakContext._return_config')
    @patch('lakshmi.cache')
    @patch('pathlib.Path.exists')
    def test_lak_context_init_with_no_config(
            self, mock_exists, mock_cache, mock_return_config):
        mock_return_config.return_value = {}
        mock_exists.return_value = True

        lakctx = lak.LakContext('unused')
        self.assertFalse(lakctx.continued)
        self.assertIsNone(lakctx.whatifs)
        self.assertIsNone(lakctx.portfolio)
        self.assertEqual(
            str(Path(lak.LakContext.DEFAULT_PORTFOLIO).expanduser()),
            lakctx.portfolio_filename)
        mock_cache.set_cache_dir.assert_not_called()

    @patch('lakshmi.lak.LakContext._return_config')
    @patch('lakshmi.cache')
    @patch('pathlib.Path.exists')
    def test_lak_context_portfolio_file_not_found(
            self, mock_exists, mock_cache, mock_return_config):
        mock_return_config.return_value = {
            'portfolio': 'portfolio.yaml'}
        mock_exists.return_value = False

        # This shouldn't raise an exception until the portfolio
        # is actually loaded.
        lakctx = lak.LakContext('unused')

        with self.assertRaisesRegex(
                click.ClickException,
                'Portfolio file portfolio.yaml does not'):
            lakctx.get_portfolio()

        mock_cache.set_cache_dir.assert_not_called()
        mock_exists.assert_called_with()

    @patch('lakshmi.lak.LakContext._return_config')
    @patch('builtins.open')
    def test_lak_context_performance_file_not_found(
            self, mock_open, mock_return_config):
        mock_return_config.return_value = {
            'performance': 'performance.yaml'}
        mock_open.side_effect = FileNotFoundError('Not found (unused)')

        lakctx = lak.LakContext('unused')
        with self.assertRaisesRegex(
                click.ClickException,
                'Performance file performance.yaml not found.'):
            lakctx.get_performance()

        mock_open.assert_called_once()

    def test_list_total(self):
        result = run_lak('list -f plain total')
        self.assertEqual(0, result.exit_code)
        self.assertIn('Total Assets  $100.00', result.output)
        self.assertNotIn('\n\n', result.output)
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_with_chaining(self):
        result = run_lak('list al total')
        self.assertEqual(0, result.exit_code)
        # Test that the separater was printed.
        self.assertIn('\n\n', result.output)
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_aa_no_args(self):
        result = run_lak('list aa')
        self.assertEqual(0, result.exit_code)
        # Check if compact version was printed.
        self.assertRegex(result.output, r'Class +A% +D%')
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_aa_no_compact(self):
        result = run_lak('list aa --no-compact')
        self.assertEqual(0, result.exit_code)
        # Check if tree version was printed.
        self.assertRegex(result.output, r'Class +Actual% .+Value\n')
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_aa_class_with_bad_args(self):
        result = run_lak('list aa --no-compact --asset-class a,b,c')
        self.assertEqual(2, result.exit_code)
        self.assertTrue('is only supported' in result.output)
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_aa_class(self):
        result = run_lak('list aa --asset-class Stocks,Bonds')
        self.assertEqual(0, result.exit_code)
        # Check if correct version was printed.
        self.assertRegex(result.output, r'Class +Actual% .+Difference\n')
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_assets(self):
        result = run_lak('list assets')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Account +Asset +Value\n')
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_assets_no_names(self):
        result = run_lak('list assets --no-long-name')
        self.assertEqual(1, result.exit_code)

    def test_list_accounts(self):
        result = run_lak('list accounts')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Account +Account Type +Value')
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_lots(self):
        result = run_lak('list lots')
        self.assertEqual(0, result.exit_code)
        self.assertEqual('', result.output)
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_list_checkpoints_no_dates(self):
        result = run_lak('list checkpoints')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'2021/01/01 +\$100.00',)
        self.assertFalse(lak.lakctx.saved_performance)

    def test_list_checkpoints_with_date(self):
        result = run_lak('list checkpoints -b 2021/01/02')
        self.assertEqual(0, result.exit_code)
        self.assertNotRegex(result.output, r'2021/01/01 +\$100.00',)
        self.assertRegex(result.output, r'2021/01/02 +\$105.01',)
        self.assertFalse(lak.lakctx.saved_performance)

    @patch('lakshmi.lak._get_todays_checkpoint')
    @patch('lakshmi.lak._today')
    def test_list_performance(self, mock_today, mock_today_checkpoint):
        mock_today.return_value = '2021/01/10'
        mock_today_checkpoint.return_value = Checkpoint('2021/01/10', 200)
        result = run_lak('list performance')

        mock_today.assert_called_once()
        mock_today_checkpoint.assert_called_once()
        self.assertEqual(0, result.exit_code)
        # Check if $10 inflow is there and +$100.00 portfolio change is also
        # in the output (because of today's checkpoint).
        self.assertRegex(result.output, r'Overall +\$10.00.+\+\$100.00',)
        self.assertFalse(lak.lakctx.saved_performance)
        self.assertFalse(lak.lakctx.performance.get_timeline().has_checkpoint(
            '2021/01/10'))

    def test_list_what_ifs_empty(self):
        result = run_lak('list whatifs')
        self.assertEqual(0, result.exit_code)
        self.assertEqual('', result.output)
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_what_if(self):
        result = run_lak('whatif asset -a Test -100')
        self.assertEqual(0, result.exit_code)
        self.assertEqual('', result.output)
        self.assertTrue(lak.lakctx.saved_portfolio)
        lak.lakctx.reset()

        result = run_lak('list whatifs')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Account +Cash\n')
        self.assertRegex(result.output, r'Account +Asset +Delta\n')
        self.assertFalse(lak.lakctx.saved_portfolio)
        lak.lakctx.reset()

        result = run_lak('list assets')
        self.assertEqual(0, result.exit_code)
        self.assertIn('Hypothetical what ifs', result.output)
        self.assertFalse(lak.lakctx.saved_portfolio)
        lak.lakctx.reset()

        result = run_lak('whatif --reset')
        self.assertEqual(0, result.exit_code)
        self.assertEqual('', result.output)
        self.assertTrue(lak.lakctx.saved_portfolio)
        lak.lakctx.reset()

        result = run_lak('list whatifs')
        self.assertEqual(0, result.exit_code)
        self.assertEqual('', result.output)
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_what_if_forced(self):
        result = run_lak('whatif asset -a Test -100')
        self.assertEqual(0, result.exit_code)
        self.assertEqual('', result.output)
        lak.lakctx.reset()

        result = run_lak('list whatifs -s')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Account +Cash\n')
        self.assertRegex(result.output, r'Account +Name +Asset +Delta\n')
        lak.lakctx.reset()

        result = run_lak('list assets whatifs -s')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Account +Cash\n')
        self.assertRegex(result.output, r'Account +Name +Asset +Delta\n')
        self.assertFalse(lak.lakctx.saved_portfolio)
        lak.lakctx.reset()

    def test_what_if_account(self):
        result = run_lak('whatif account -t Schwab -100')
        self.assertEqual(0, result.exit_code)
        self.assertEqual('', result.output)
        self.assertTrue(lak.lakctx.saved_portfolio)
        lak.lakctx.reset()

        result = run_lak('list whatifs')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Account +Cash\n')
        self.assertNotRegex(result.output, r'Account +Asset +Delta\n')
        self.assertFalse(lak.lakctx.saved_portfolio)
        lak.lakctx.reset()

    def test_info_account(self):
        result = run_lak('info account -t Schwab')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Name: +Schwab\n')
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_info_asset(self):
        result = run_lak('info asset -a Test')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Name: +Test Asset\n')
        self.assertFalse(lak.lakctx.saved_portfolio)

    @patch('lakshmi.lak._get_todays_checkpoint')
    @patch('lakshmi.lak._today')
    def test_info_performance(self, mock_today, mock_today_checkpoint):
        mock_today.return_value = '2021/01/10'
        mock_today_checkpoint.return_value = Checkpoint('2021/01/10', 200)
        result = run_lak('info performance --begin 2021/1/1')

        mock_today.assert_called_once()
        mock_today_checkpoint.assert_called_once()
        self.assertEqual(0, result.exit_code)
        self.assertRegex(
            result.output,
            r'Start date +2021/01/01\nEnd date +2021/01/10\n')
        self.assertFalse(lak.lakctx.saved_portfolio)
        self.assertFalse(lak.lakctx.performance.get_timeline().has_checkpoint(
            '2021/01/10'))

    @patch('click.edit')
    @patch('pathlib.Path.read_text')
    def test_edit_and_parse_with_no_dict(self, mock_read_text, mock_edit):
        mock_read_text.return_value = 'a: b'
        mock_edit.return_value = 'c: d'

        actual = lak.edit_and_parse(None, lambda x: x, 'test_file')

        self.assertEqual({'c': 'd'}, actual)
        mock_read_text.assert_called_once()
        mock_edit.assert_called_with('a: b')

    @patch('click.edit')
    @patch('pathlib.Path.read_text')
    def test_edit_and_parse_with_dict(self, mock_read_text, mock_edit):
        mock_read_text.return_value = 'a: b'
        mock_edit.return_value = 'c: d\n\n' + lak._HELP_MSG_PREFIX

        actual = lak.edit_and_parse({'e': 'f'}, lambda x: x, 'test_file')

        self.assertEqual({'c': 'd'}, actual)
        mock_read_text.assert_called_once()
        mock_edit.assert_called_with(
            'e: f\n' + lak._HELP_MSG_PREFIX + '# a: b')

    @patch('click.edit')
    @patch('pathlib.Path.read_text')
    def test_edit_and_parse_with_comma_floats(self, mock_read_text, mock_edit):
        mock_read_text.return_value = 'a: b'
        mock_edit.return_value = 'c: 123,456.78\n\n' + lak._HELP_MSG_PREFIX

        actual = lak.edit_and_parse({'e': 'f'}, lambda x: x, 'test_file')

        self.assertEqual({'c': 123456.78}, actual)
        mock_read_text.assert_called_once()
        mock_edit.assert_called_with(
            'e: f\n' + lak._HELP_MSG_PREFIX + '# a: b')

    @patch('click.edit')
    @patch('pathlib.Path.read_text')
    def test_edit_and_parse_aborted(self, mock_read_text, mock_edit):
        mock_read_text.return_value = 'a: b'
        mock_edit.return_value = None

        with self.assertRaises(click.Abort):
            lak.edit_and_parse(None, lambda x: x, 'test_file')

        mock_read_text.assert_called_once()
        mock_edit.assert_called_with('a: b')

    @patch('click.echo')
    @patch('click.confirm')
    @patch('click.edit')
    @patch('pathlib.Path.read_text')
    def test_edit_and_parse_user_aborted(self, mock_read_text, mock_edit,
                                         mock_confirm, mock_echo):
        mock_read_text.return_value = 'a: b'
        mock_edit.return_value = 'c: d'
        mock_confirm.return_value = False

        def parse_fn(x):
            raise Exception('Better luck next time')

        with self.assertRaises(click.Abort):
            lak.edit_and_parse(None, parse_fn, 'test_file')

        mock_read_text.assert_called_once()
        mock_edit.assert_called_with('a: b')
        mock_confirm.assert_called_once()
        mock_echo.assert_called_with('Error parsing file: '
                                     "Exception('Better luck next time')")

    @patch('click.echo')
    @patch('click.confirm')
    @patch('click.edit')
    @patch('pathlib.Path.read_text')
    def test_edit_and_parse_user_fixed(self, mock_read_text, mock_edit,
                                       mock_confirm, mock_echo):
        mock_read_text.return_value = 'a: b'
        mock_edit.side_effect = ['c~~d', 'c: d']
        mock_confirm.return_value = True

        def parse_fn(x):
            if x == 'c~~d':
                raise Exception('Better luck next time')
            else:
                return x

        actual = lak.edit_and_parse(None, parse_fn, 'test_file')
        self.assertEqual({'c': 'd'}, actual)

        mock_read_text.assert_called_once()
        mock_edit.assert_has_calls([unittest.mock.call('a: b'),
                                    unittest.mock.call('c~~d')])
        mock_confirm.assert_called_once()
        mock_echo.assert_called_with('Error parsing file: '
                                     "Exception('Better luck next time')")

    @patch('pathlib.Path.exists')
    def test_init_portfolio_exists(self, mock_exists):
        mock_exists.return_value = True

        result = run_lak('init')
        self.assertEqual(1, result.exit_code)
        self.assertIn('Portfolio file already', result.output)
        self.assertFalse(lak.lakctx.saved_portfolio)

    @patch('lakshmi.lak.edit_and_parse')
    @patch('pathlib.Path.exists')
    def test_init_portfolio(self, mock_exists, mock_parse):
        mock_exists.return_value = False
        mock_parse.return_value = AssetClass('Money')

        result = run_lak('init')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)
        self.assertEqual('Money', lak.lakctx.portfolio.asset_classes.name)

    @patch('lakshmi.lak.edit_and_parse')
    def test_edit_asset_class(self, mock_parse):
        mock_parse.return_value = AssetClass('Money')

        previous_ac_dict = lak.lakctx.portfolio.asset_classes.to_dict()
        result = run_lak('edit assetclass')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)
        self.assertEqual('Money', lak.lakctx.portfolio.asset_classes.name)

        mock_parse.assert_called_with(previous_ac_dict,
                                      unittest.mock.ANY,
                                      'AssetClass.yaml')

    def test_edit_account_bad_name(self):
        result = run_lak('edit account -t Yolo')
        self.assertEqual(1, result.exit_code)
        self.assertFalse(lak.lakctx.saved_portfolio)

    @patch('lakshmi.lak.edit_and_parse')
    def test_edit_account_change_type(self, mock_parse):
        mock_parse.return_value = Account('Schwab', 'Tax-exempt')

        result = run_lak('edit account -t Schwab')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)

        accounts = list(lak.lakctx.portfolio.accounts())
        self.assertEqual(1, len(accounts))
        self.assertEqual('Tax-exempt', accounts[0].account_type)
        self.assertEqual(1, len(accounts[0].assets()))

        mock_parse.assert_called_with(Account('Schwab', 'Taxable').to_dict(),
                                      unittest.mock.ANY,
                                      'Account.yaml')

    @patch('lakshmi.lak.edit_and_parse')
    def test_edit_account_change_name(self, mock_parse):
        mock_parse.return_value = Account('Vanguard', 'Taxable')

        result = run_lak('edit account -t Schwab')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)

        accounts = list(lak.lakctx.portfolio.accounts())
        self.assertEqual(1, len(accounts))
        self.assertEqual('Vanguard', accounts[0].name())
        self.assertEqual(1, len(accounts[0].assets()))

        mock_parse.assert_called_with(Account('Schwab', 'Taxable').to_dict(),
                                      unittest.mock.ANY,
                                      'Account.yaml')

    @patch('lakshmi.lak.edit_and_parse')
    def test_edit_asset(self, mock_parse):
        mock_parse.return_value = ManualAsset(
            'Tasty Asset', 100.0, {'Stocks': 1.0})

        result = run_lak('edit asset -a Test')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)

        account = lak.lakctx.portfolio.get_account('Schwab')
        self.assertEqual(1, len(account.assets()))
        self.assertEqual('Tasty Asset', list(account.assets())[0].name())

        mock_parse.assert_called_with(
            ManualAsset('Test Asset', 100.0, {'Stocks': 1.0}).to_dict(),
            unittest.mock.ANY,
            'ManualAsset.yaml')

    @patch('lakshmi.lak.edit_and_parse')
    def test_edit_checkpoint(self, mock_parse):
        mock_parse.return_value = Checkpoint('2021/1/2', 105.01,
                                             inflow=10, outflow=1)
        result = run_lak('edit checkpoint --date 2021/01/02')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_performance)
        self.assertEqual(
            1, lak.lakctx.get_performance().get_timeline().get_checkpoint(
                '2021/01/02').get_outflow())

    @patch('lakshmi.lak.edit_and_parse')
    def test_add_account(self, mock_parse):
        mock_parse.return_value = Account('Vanguard', 'Taxable')

        result = run_lak('add account')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)

        self.assertEqual(2, len(lak.lakctx.portfolio.accounts()))
        mock_parse.assert_called_with(None,
                                      Account.from_dict,
                                      'Account.yaml')

    @patch('lakshmi.lak.edit_and_parse')
    def test_add_asset(self, mock_parse):
        mock_parse.return_value = ManualAsset(
            'Tasty Asset', 100.0, {'Stocks': 1.0})

        result = run_lak('add asset -t Schwab -p ManualAsset')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)

        account = lak.lakctx.portfolio.get_account('Schwab')
        self.assertEqual(2, len(account.assets()))

        mock_parse.assert_called_with(
            None,
            ManualAsset.from_dict,
            'ManualAsset.yaml')

    @patch('lakshmi.lak._today')
    def test_add_checkpoint(self, mock_today):
        mock_today.return_value = '2021/01/31'

        result = run_lak('add checkpoint')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_performance)
        self.assertEqual(
            100.0,
            lak.lakctx.get_performance().get_timeline().get_checkpoint(
                '2021/01/31').get_portfolio_value())

    @patch('lakshmi.lak._today')
    def test_add_checkpoint_to_empty(self, mock_today):
        mock_today.return_value = '2100/01/31'
        lak.lakctx.performance = None

        result = run_lak('add checkpoint')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_performance)
        self.assertEqual('2100/01/31',
                         lak.lakctx.get_performance().get_timeline().begin())

    @patch('lakshmi.lak.edit_and_parse')
    @patch('lakshmi.lak._today')
    def test_add_checkpoint_and_edit(self, mock_today, mock_parse):
        mock_today.return_value = '2021/01/31'
        mock_parse.return_value = Checkpoint('2021/01/31', 500.0)

        result = run_lak('add checkpoint --edit')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_performance)
        self.assertEqual(
            500.0,
            lak.lakctx.get_performance().get_timeline().get_checkpoint(
                '2021/01/31').get_portfolio_value())

    def test_delete_account(self):
        result = run_lak('delete account -t Schwab --yes')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)
        self.assertEqual(0, len(lak.lakctx.portfolio.accounts()))

    def test_delete_asset(self):
        result = run_lak('delete asset -a Test --yes')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_portfolio)
        self.assertEqual(
            0, len(lak.lakctx.portfolio.get_account('Schwab').assets()))

    def test_delete_checkpoint(self):
        result = run_lak('delete checkpoint --date 2021/1/1 --yes')
        self.assertEqual(0, result.exit_code)
        self.assertTrue(lak.lakctx.saved_performance)
        self.assertEqual('2021/01/02',
                         lak.lakctx.get_performance().get_timeline().begin())

    def test_analyze_tlh(self):
        result = run_lak('analyze tlh')
        self.assertEqual(0, result.exit_code)
        self.assertIn('No tax lots', result.output)

    def test_analyze_rebalance(self):
        result = run_lak('analyze rebalance')
        self.assertEqual(0, result.exit_code)
        self.assertRegex(result.output, r'Bonds +0')

    def test_analyze_allocate_no_cash(self):
        result = run_lak('analyze allocate -t Schwab')
        self.assertEqual(1, result.exit_code)
        self.assertIn('No available cash', str(result.exception))
        self.assertFalse(lak.lakctx.saved_portfolio)

    def test_analyze_allocate(self):
        self.assertEqual(0, run_lak('whatif account -t Schwab 100').exit_code)
        result = run_lak('analyze allocate -t Schwab')
        self.assertEqual(0, result.exit_code)
        self.assertIn('+$100.00', result.output)
        self.assertTrue(lak.lakctx.saved_portfolio)
        self.assertEqual(0, run_lak('whatif -r').exit_code)


if __name__ == '__main__':
    unittest.main()
